// Copyright: Daisuke Okanohara
// Lincense: New BSD Lincense

#include "tx.hpp"
#include "ssv.hpp"

namespace tx_tool{
  uint tx::NOTFOUND = UINT_MAX;

  struct queue_elem{
    queue_elem(size_t _left, size_t _right, int _depth) :left(_left),right(_right),depth(_depth){}
    size_t left;
    size_t right;
    int depth;
  };


  tx::tx():edge(NULL){}

  tx::~tx(){
    if (edge) delete[] edge;
  }

  int tx::build(vector<string>& wordList, const char* fileName) const{
    sort(wordList.begin(),wordList.end());
    const size_t origElemNum = wordList.size();
    wordList.erase(unique(wordList.begin(),wordList.end()),wordList.end());
    const int elemNum = (int)wordList.size();
    if (elemNum != origElemNum){
      // fprintf(stderr,"shrink word list %d -> %d\n",origElemNum,elemNum);
    } else {
      // fprintf(stderr,"word list %d elements\n",elemNum);
    }

    uint totalSize = 0;
    for (size_t i = 0; i < wordList.size(); i++){
      totalSize += (uint)wordList[i].size();
    }
        
        
    FILE* outfp = fopen(fileName,"wb");
    if (outfp == NULL){
      // fprintf(stderr,"cannot open %s\n",fileName);
      return -1;
    }
        
    queue<queue_elem> q;
    q.push(queue_elem(0,elemNum,0));
    fwrite(&elemNum,sizeof(int),1,outfp);

    vector<bool> vb_loud;
    vector<bool> vb_terminal;

    vb_loud.push_back(0); // super root
    vb_loud.push_back(1); 

    uint nodeNum = 0;
    while (!q.empty()){
      queue_elem& elem = q.front();
      const int depth    = elem.depth;
      const size_t left  = elem.left;
      const size_t right = elem.right;
      q.pop();

      nodeNum++;
      size_t newLeft = left;
      if (wordList[left].size() == depth){
        vb_terminal.push_back(1); // this node has terminate
        newLeft++;
        if (newLeft == right){
          vb_loud.push_back(1);
          continue;
        }
      } else {
        vb_terminal.push_back(0);
      }
      size_t prev = newLeft;
      char prev_c = wordList[prev][depth];
      for (size_t i = newLeft+1; i < right; i++){
        if (prev_c != wordList[i][depth]){
          fputc(prev_c,outfp);
          vb_loud.push_back(0);
          q.push(queue_elem(prev,i,depth+1));
          prev = i;
          prev_c = wordList[prev][depth];
        }
      }
      if (prev != right){
        fputc(prev_c,outfp);
        vb_loud.push_back(0);
        q.push(queue_elem(prev,right,depth+1));
      }
      vb_loud.push_back(1);
    }

    {
      ssv sv(vb_loud);
      sv.build();
      sv.write(outfp);
    }
    {
      ssv sv(vb_terminal);
      sv.build();
      sv.write(outfp);
    }
        
    fwrite(&nodeNum,sizeof(int),1,outfp);

    // size_t outfpSize = ftell(outfp);
    // fprintf(stderr,"outputSize:%d inputSize:%d ratio:%.2f\n",outfpSize,totalSize,(float)outfpSize*100.f/totalSize);
    if (outfp) fclose(outfp);
    return 0;
  }

  int tx::read(const char* fileName){
    FILE* infp = fopen(fileName,"rb");
    if (infp == NULL){
      // fprintf(stderr,"cannot open %s\n",fileName);
      return -1;
    }

    int elemNum = -1;
    if (fread(&elemNum,sizeof(int),1,infp) != 1){
      // fprintf(stderr,"elemNum read error\n");
      fclose(infp);
      return -1;
    }

    fseek(infp,0,SEEK_END);
    size_t fileSize = ftell(infp);
    if (fseek(infp,fileSize-(1*sizeof(int)),SEEK_SET) == -1){
      // fprintf(stderr,"fseek error\n");
      fclose(infp);
      return -1;
    }

    int nodeNum = -1;
    if (fread(&nodeNum,sizeof(int),1,infp) != 1){
      // fprintf(stderr,"nodeNum read error\n");
      fclose(infp);
      return -1;
    }
    // fprintf(stderr,"elemNum:%d nodeNum:%d\n",elemNum,nodeNum);

    if (fseek(infp,sizeof(int)*1,SEEK_SET) == -1){
      // fprintf(stderr,"fseek error\n");
      fclose(infp);
      return -1;
    }

    edge = new char [nodeNum-1];
    if (fread(edge,sizeof(char),nodeNum-1,infp) != nodeNum-1){
      // fprintf(stderr,"fseek error\n");
      fclose(infp);
      return -1;
    }

    loud.read(infp);
    terminal.read(infp);

    if (infp) fclose(infp);
    return 0;
  }
  
  uint tx::searchLongestPrefix(const char* str, const size_t len, size_t& retLen, bool matchPrefix) const {
    uint curPos = 2;
    uint retId = NOTFOUND;
    for (size_t i = 0 ; ; i++){
      const uint nodeId = loud.rank(curPos-1,1)-1;
      if (matchPrefix || terminal.getBit(nodeId)){
        retLen = i;
        retId = terminal.rank(nodeId,1)-1;
      }
      if (i == len) break;
      uint nextPos = getChild(curPos,str[i]);
      if (nextPos == UINT_MAX){
        break;
      }
      curPos = nextPos;
    }
    return retId;
  }

  uint tx::searchPrefixes(const char* str, const size_t len, vector<string>& ret) const {
    strToPos(str, len, &ret);
    return (uint)ret.size();
  }
  
  uint tx::searchExpansions(const char* str, const size_t len, vector<string>& ret, const uint limit) const {    
    uint curPos = strToPos(str, len, NULL);
    if (curPos != UINT_MAX){
      string curStr(str,len);
      vector<pair<size_t,string> > ret_p;
      enumerateAll(curPos,curStr,ret_p); 
      sort(ret_p.begin(),ret_p.end());
      const size_t newLimit = (limit == 0) ? ret_p.size() : limit; 
      for (size_t i = 0; i < ret_p.size()&& i < newLimit; i++){
        ret.push_back(ret_p[i].second);
      }
    } 
    return (uint)ret.size();
  }

  void tx::enumerateAll(const uint pos, const string str, vector<pair<size_t,string> >& ret) const{
    uint curPos = pos;
    uint edgePos = loud.rank(pos,0)-2;
    while (loud.getBit(curPos) == 0){
      const uint nextPos = loud.select(loud.rank(curPos,0),1)+1;
      string childStr = str + edge[edgePos];
      if (isTerminal(nextPos)){
        ret.push_back(make_pair<size_t,string>(childStr.size(),childStr));
      }
      enumerateAll(nextPos,childStr,ret);
      curPos++;
      edgePos++;
    }
  }

  uint tx::getChild(const uint pos, const char c) const{
    uint curPos = pos;
    uint edgePos = loud.rank(pos,0)-2;
    for (;;){
      if (loud.getBit(curPos) == 1) {
        curPos = UINT_MAX;
        return curPos;
      }
      if (edge[edgePos] == c){
        uint nextPos = loud.select(loud.rank(curPos,0),1)+1;
        return nextPos;
      }
      curPos++;
      edgePos++;
    }
  }
  
  uint tx::strToPos(const char* str, const size_t len, vector<string>* prefixes) const {
    uint curPos = 2;
    for (size_t i = 0; i <= len; i++){
      if (isTerminal(curPos)){
        if (prefixes) prefixes->push_back(string(str,str+i));
      }
      if (i == len) break;
      curPos = getChild(curPos,str[i]);
      if (curPos == UINT_MAX) break;
    }
    return curPos;
  }
  
  bool tx::isTerminal(const uint pos) const{
    const uint nodeId = loud.rank(pos-1,1)-1;
    return terminal.getBit(nodeId);
  }

} // namespace tx_tool
